searchNodes=[{"doc":"A library to serialize between maps and nested structs. Turns a map into a nested struct according to hints given to the library. And vice versa turns any nested struct into a map. It works with maps/structs of any shape and level of nesting. Highly configurable by implementing Nestru.Decoder and Nestru.Encoder protocols for structs. Useful for translating map keys to struct's fields named differently. Or to specify default values missing in the map and required by struct. The library's primary purpose is to serialize a map coming from a JSON payload or an Erlang term; at the same time, the map can be of any origin. The input map can have atom or binary keys. The library takes the binary key first and then the same-named atom key if the binary key is missing while decoding the map. The library generates maps with atom keys during the struct encode operation. Tour Let's say we have an Order with a total field which is an instance of a Total struct. And we want to serialize between an instance of Order and a map. Firstly, let's derive Nestru.Encoder and Nestru.Decoder protocols and give a hint that the field :total should hold a value of Total struct like the following: defmodule Order do @derive [ Nestru.Encoder , { Nestru.Decoder , hint : %{ total : Total } } ] defstruct [ :id , :total ] end defmodule Total do @derive [ Nestru.Encoder , Nestru.Decoder ] defstruct [ :sum ] end {:module, Total, &lt;&lt;70, 79, 82, 49, 0, 0, 8, ...&gt;&gt;, %Total{sum: nil}} Secondly, we can encode the Order into the map like that: model = % Order { id : &quot;A548&quot; , total : % Total { sum : 500 } } { :ok , map } = Nestru . encode ( model ) {:ok, %{id: &quot;A548&quot;, total: %{sum: 500}}} And decode the map back into the Order like the following: map = %{ &quot;id&quot; =&gt; &quot;A548&quot; , &quot;total&quot; =&gt; %{ &quot;sum&quot; =&gt; 500 } } { :ok , model } = Nestru . decode ( map , Order ) {:ok, %Order{id: &quot;A548&quot;, total: %Total{sum: 500}}} As you can see the data markup is in place, the Total struct is nested within the Order struct. A list of structs in a field Let's add the :items field to Order1 struct to hold a list of LineItem s and give a hint to Nestru on how to decode that field: defmodule Order1 do @derive { Nestru.Decoder , hint : %{ total : Total , items : [ LineItem ] } } defstruct [ :id , :items , :total ] end defmodule LineItem do @derive Nestru.Decoder defstruct [ :amount ] end {:module, LineItem, &lt;&lt;70, 79, 82, 49, 0, 0, 8, ...&gt;&gt;, %LineItem{amount: nil}} Let's decode: map = %{ &quot;id&quot; =&gt; &quot;A548&quot; , &quot;items&quot; =&gt; [ %{ &quot;amount&quot; =&gt; 150 } , %{ &quot;amount&quot; =&gt; 350 } ] , &quot;total&quot; =&gt; %{ &quot;sum&quot; =&gt; 500 } } { :ok , model } = Nestru . decode ( map , Order1 ) {:ok, %Order1{ id: &quot;A548&quot;, items: [%LineItem{amount: 150}, %LineItem{amount: 350}], total: %Total{sum: 500} }} Voilà, we have field values as nested structs 🎉 For the case when the list contains several structs of different types, please, see the Serializing type-dependent fields section below. Date Time and URI Let's say we have an Order2 struct with some URI and DateTime fields in it. These attributes are structs in Elixir, at the same time they usually kept as binary representations in a map. Nestru supports conversion between binaries and structs, all we need to do is to implement the Nestry.Encoder and Nestru.Decoder protocols for these structs like the following: # DateTime defimpl Nestru.Encoder , for : DateTime do def gather_fields_from_struct ( struct , _context ) do { :ok , DateTime . to_string ( struct ) } end end defimpl Nestru.Decoder , for : DateTime do def decode_fields_hint ( _empty_struct , _context , value ) do case DateTime . from_iso8601 ( value ) do { :ok , date_time , _offset } -&gt; { :ok , date_time } error -&gt; error end end end # URI defimpl Nestru.Encoder , for : URI do def gather_fields_from_struct ( struct , _context ) do { :ok , URI . to_string ( struct ) } end end defimpl Nestru.Decoder , for : URI do def decode_fields_hint ( _empty_struct , _context , value ) do URI . new ( value ) end end {:module, Nestru.Decoder.URI, &lt;&lt;70, 79, 82, 49, 0, 0, 8, ...&gt;&gt;, {:decode_fields_hint, 3}} Order2 is defined like this: defmodule Order2 do @derive [ Nestru.Encoder , { Nestru.Decoder , hint : %{ date : DateTime , website : URI } } ] defstruct [ :id , :date , :website ] end {:module, Order2, &lt;&lt;70, 79, 82, 49, 0, 0, 8, ...&gt;&gt;, %Order2{id: nil, date: nil, website: nil}} We can encode it to a map with binary fields like the following: order = % Order2 { id : &quot;B445&quot; , date : ~U[2024-03-15 22:42:03Z] , website : URI . parse ( &quot;https://www.example.com/?book=branch&quot; ) } { :ok , map } = Nestru . encode ( order ) {:ok, %{id: &quot;B445&quot;, date: &quot;2024-03-15 22:42:03Z&quot;, website: &quot;https://www.example.com/?book=branch&quot;}} And decode it back: Nestru . decode ( map , Order2 ) {:ok, %Order2{ id: &quot;B445&quot;, date: ~U[2024-03-15 22:42:03Z], website: %URI{ scheme: &quot;https&quot;, userinfo: nil, host: &quot;www.example.com&quot;, port: 443, path: &quot;/&quot;, query: &quot;book=branch&quot;, fragment: nil } }} Error handling and path to the failed part of the map Every implemented function of Nestru protocols can return {error, message} tuple in case of failure. When Nestru receives the error tuple, it stops conversion and bypasses the error to the caller. defmodule Location do @derive { Nestru.Decoder , hint : %{ street : Street } } defstruct [ :street ] end defmodule Street do @derive { Nestru.Decoder , hint : %{ house : House } } defstruct [ :house ] end defmodule House do defstruct [ :number ] defimpl Nestru.Decoder do def decode_fields_hint ( _empty_struct , _context , value ) do if Nestru . has_key? ( value , :number ) do { :ok , %{ } } else { :error , &quot;Can&#39;t continue without house number.&quot; } end end end end So when we decode the following map missing the number value, we will get the error back: map = %{ &quot;street&quot; =&gt; %{ &quot;house&quot; =&gt; %{ &quot;name&quot; =&gt; &quot;Party house&quot; } } } { :error , error } = Nestru . decode ( map , Location ) {:error, %{ get_in_keys: [#Function&lt;8.67001686/3 in Access.key!/1&gt;, #Function&lt;8.67001686/3 in Access.key!/1&gt;], message: &quot;Can't continue without house number.&quot;, path: [&quot;street&quot;, &quot;house&quot;] }} Nestru wraps the error message into a map and adds path and get_in_keys fields to it. The path values point to the failed part of the map which can be returned like the following: get_in ( map , error . get_in_keys ) %{&quot;name&quot; =&gt; &quot;Party house&quot;} Maps with different key names In some cases, the map's keys have slightly different names compared to the target's struct field names. Fields that should be decoded into the struct can be gathered by adopting Nestru.PreDecoder protocol like the following: defmodule Quote do @derive [ { Nestru.PreDecoder , translate : %{ &quot;cost_value&quot; =&gt; :cost } } , Nestru.Decoder ] defstruct [ :cost ] end When we decode the map, Nestru will put the value of the &quot;cost_value&quot; key for the :cost key into the map and then complete the decoding: map = %{ &quot;cost_value&quot; =&gt; 1280 } Nestru . decode ( map , Quote ) {:ok, %Quote{cost: 1280}} For more sophisticated key mapping you can implement the gather_fields_for_decoding/3 function of Nestru.PreDecoder explicitly. Serializing type-dependent fields To convert a struct with a field that can have the value of multiple struct types into the map and back, the type of the field's value should be persisted. It's possible to do that like the following: defmodule BookCollection do defstruct [ :name , :items ] defimpl Nestru.Encoder do def gather_fields_from_struct ( struct , _context ) do items_kinds = Enum . map ( struct . items , fn % module { } -&gt; module |&gt; Module . split ( ) |&gt; Enum . join ( &quot;.&quot; ) end ) { :ok , %{ name : struct . name , items : struct . items , items_kinds : items_kinds } } end end defimpl Nestru.Decoder do def decode_fields_hint ( _empty_struct , _context , value ) do items_kinds = Enum . map ( value . items_kinds , fn module_string -&gt; module_string |&gt; String . split ( &quot;.&quot; ) |&gt; Module . safe_concat ( ) end ) { :ok , %{ items : &amp; Nestru . decode_from_list ( &amp;1 , items_kinds ) } } end end end defmodule BookCollection.Book do @derive [ Nestru.Encoder , Nestru.Decoder ] defstruct [ :title ] end defmodule BookCollection.Magazine do @derive [ Nestru.Encoder , Nestru.Decoder ] defstruct [ :issue ] end Let's convert the nested struct into a map. The returned map gets extra items_kinds field with types information: alias BookCollection . { Book , Magazine } collection = % BookCollection { name : &quot;Duke of Norfolk&#39;s archive&quot; , items : [ % Book { title : &quot;The Spell in the Chasm&quot; } , % Magazine { issue : &quot;Strange Hunt&quot; } ] } { :ok , map } = Nestru . encode ( collection ) {:ok, %{ items: [%{title: &quot;The Spell in the Chasm&quot;}, %{issue: &quot;Strange Hunt&quot;}], items_kinds: [&quot;BookCollection.Book&quot;, &quot;BookCollection.Magazine&quot;], name: &quot;Duke of Norfolk's archive&quot; }} And restoring of the original nested struct is as simple as that: { :ok , collection } = Nestru . decode ( map , BookCollection ) {:ok, %BookCollection{ items: [ %BookCollection.Book{title: &quot;The Spell in the Chasm&quot;}, %BookCollection.Magazine{issue: &quot;Strange Hunt&quot;} ], name: &quot;Duke of Norfolk's archive&quot; }} Use with other libraries Jason JSON maps decoded with Jason library are supported with both binary and atoms keys. ex_json_schema ex_json_schema library can be used before decoding the input map with the JSON schema. To make sure that the structure of the input map is correct. ExJSONPath ExJsonPath library allows querying maps (JSON objects) and lists (JSON arrays), using JSONPath expressions. The queries can be useful in Nestru.PreDecoder.gather_fields_for_decoding/3 function to assemble fields for decoding from a map having a very different shape from the target struct. Domo You can use the Domo library to validate the t() types of the nested struct values after decoding with Nestru . Domo can validate a nested struct in one pass, ensuring that the struct's field values match its t() type and associated preconditions.","ref":"Nestru.html","title":"Nestru","type":"module"},{"doc":"Decodes a map or a binary into the given struct. The first argument is a map having key-value pairs which supports both string and atom keys. Or a binary representation, f.e. date time in ISO 8601 format. The second argument is a struct's module atom. The third argument is a context value to be passed to implemented functions of Nestru.PreDecoder and Nestru.Decoder protocols. To give a hint on how to decode nested struct values or a list of such values for the given field, implement Nestru.Decoder protocol for the struct. Function calls struct/2 to build the struct's value. If given a map, keys that don't exist in the struct are automatically discarded.","ref":"Nestru.html#decode/3","title":"Nestru.decode/3","type":"function"},{"doc":"Similar to decode/3 but checks if enforced struct's fields keys exist after decoding. Returns a struct or raises an error.","ref":"Nestru.html#decode!/3","title":"Nestru.decode!/3","type":"function"},{"doc":"Decodes a list of values into the list of the given structs. The first argument is an input list. If the second argument is a struct's module atom, then the function calls the decode/3 on each input list item with the given module atom as the second argument. If the second argument is a list of struct module atoms, the function calls the decode/3 function on each input list item with the module atom taken at the same index of the second list as the second argument. In this case, both lists should be of equal length. The third argument is a context value to be passed to implemented functions of Nestru.PreDecoder and Nestru.Decoder protocols. The function returns a list of structs or the first error from decode/3 function.","ref":"Nestru.html#decode_from_list/3","title":"Nestru.decode_from_list/3","type":"function"},{"doc":"Similar to decode_from_list/2 but checks if enforced struct's fields keys exist in the given maps. Returns a struct or raises an error.","ref":"Nestru.html#decode_from_list!/3","title":"Nestru.decode_from_list!/3","type":"function"},{"doc":"Encodes the given struct into a map. The first argument is a struct value to be encoded into map. Encodes each field's value recursively when it is a struct or a list of structs. The second argument is a context to be passed to Nestru.Encoder protocol function. To insert additional fields or rename or drop existing ones before encoding into map, implement Nestru.Encoder protocol for the struct. That can be used to keep additional type information for the field that can have a value of various value types.","ref":"Nestru.html#encode/2","title":"Nestru.encode/2","type":"function"},{"doc":"Similar to encode/1 . Returns a map or raises an error.","ref":"Nestru.html#encode!/2","title":"Nestru.encode!/2","type":"function"},{"doc":"Encodes the given list of structs into a list of values. Calls encode/2 for each struct in the list. The Nestru.Encoder protocol should be implemented for each struct module. The function returns a list of values or the first error from encode/2 function.","ref":"Nestru.html#encode_to_list/2","title":"Nestru.encode_to_list/2","type":"function"},{"doc":"Similar to encode_to_list/2 Returns list of maps or raises an error.","ref":"Nestru.html#encode_to_list!/2","title":"Nestru.encode_to_list!/2","type":"function"},{"doc":"Gets the value for a specific key in map. Lookups a binary then an atom key. If key is present in map then its value value is returned. Otherwise, default is returned. If default is not provided, nil is used.","ref":"Nestru.html#get/3","title":"Nestru.get/3","type":"function"},{"doc":"Returns whether the given key exists in the given map as a binary or as an atom.","ref":"Nestru.html#has_key?/2","title":"Nestru.has_key?/2","type":"function"},{"doc":"","ref":"Nestru.Decoder.html","title":"Nestru.Decoder","type":"protocol"},{"doc":"Returns the hint of how to decode the struct fields. The first argument is an empty struct value adopting the protocol. The second argument is the context value given to the Nestru.decode/3 function call. The third argument is a map or a binary to be decoded into the struct. If the function returns {:ok, hint} , then the hint is expected to be a map or an instance of the decodable struct. When the map is returned then, key-value pairs give the decoding hint about the struct's fields with the key equal to field's name and the value specifying the following: A module's atom (f.e. Module ) specifies that the appropriate field's value should be decoded as a nested struct defined in the module. Each field of the nested struct will be decoded recursively. A one-element list with the module's atom (f.e. [Module] ) specifies that the appropriate field's value should be decoded as a list of structs. It's equivalent of returning &amp;Nestru.decode_from_list(&amp;1, module) . An anonymous function with arity 1 (f.e. &amp;my_fun/1 ) specifies that the appropriate field's value should be returned from the function. The function's only argument is the value from the map to be decoded, and it is expected to return {:ok, term} , {:error, %{message: term, path: list}} , or {:error, term} . Any field missing the key in the hint receives the value as-is from the input map. The %{} empty hint value defines that all fields of the struct take all values from the input map unmodified. When the returned hint is an instance of the decodable struct, then it is returned as is. If the function returns {:ok, nil} , then the decoded struct's value is nil. If the function returns {:error, message} tuple, then decoding stops, and the error is bypassed to the caller. Any other return value raises an error. To generate the implementation of the function for the given struct in short form, set the @derive module attribute to the tuple of Nestru.Decoder and the :hint option specifying the hint map to be returned. The default implementation derived for the struct pulls all values from the input map unmodified. See the examples below. Examples def Supplier do # Derive the default implementation @derive Nestru.Decoder defstruct [ :id , :name ] end def FruitBox.Fruit do # Give a hint in a compact form with deriving the protocol @derive { Nestru.Decoder , hint : %{ vitamins : [ Vitamin ] , supplier : Supplier } } defstruct [ :vitamins , :supplier ] end defmodule FruitBox do defstruct [ :items ] # Implement the function returning the hint explicitly defimpl Nestru.Decoder do def decode_fields_hint ( _empty_struct , _context , value ) do # Give a function to decode the list field as a hint, other fields are copied as is { :ok , %{ items : &amp; Nestru . decode_from_list ( &amp;1 , FruitBox.Fruit ) } } end end end","ref":"Nestru.Decoder.html#decode_fields_hint/3","title":"Nestru.Decoder.decode_fields_hint/3","type":"function"},{"doc":"","ref":"Nestru.Decoder.html#t:t/0","title":"Nestru.Decoder.t/0","type":"type"},{"doc":"","ref":"Nestru.Encoder.html","title":"Nestru.Encoder","type":"protocol"},{"doc":"Returns the fields gathered from the given struct, which can be encoded further. It can be used to rename the keys of the map, or to turn the whole struct to binary. The first argument is the encodable struct value adopting the protocol. The second argument is the context given to encode/2 or encode_to_list/2 functions. Nestru calls this function as the first step of the encoding procedure. If the function returns {:ok, item} , then the encoding continues depending on the value of the item as follows: for a binary, the encoding finishes, and the binary is returned for a map, the encoding continues for each key-value pair. If the value is a struct then it is encoded into a map with encode/1 , the same is done recursively for lists, all other values are left as they are. If the function returns {:error, message} tuple, then encoding stops, and the error is bypassed to the caller. Any other return value raises an error. To generate the default implementation of the function, add the @derive Nestru.Encoder attribute to the struct. The :only and :except options are supported to filter the fields. The default implementation gathers keys from the struct by calling Map.from_struct/1 . Examples def Supplier do # Derive the default implementation @derive Nestru.Encoder defstruct [ :id , :name ] end def Purchaser do # Encode only one field @derive { Nestru.Encoder , only : [ :id ] } defstruct [ :id , :name , :address ] end defmodule BoxLabel do defstruct [ :prefix , :number ] # Encode label as a binary defimpl Nestru.Encoder do def gather_fields_from_struct ( struct , _context ) do { :ok , &quot;FF{struct.prefix}-{struct.number}&quot; } end end end defmodule FruitBox do defstruct [ :items , :label ] # Rename the :items key to :elements in the result map defimpl Nestru.Encoder do def gather_fields_from_struct ( struct , _context ) do map = Map . from_struct ( struct ) { :ok , map |&gt; Map . put ( :elements , Map . get ( map , :items ) ) |&gt; Map . delete ( :items ) } end end end","ref":"Nestru.Encoder.html#gather_fields_from_struct/2","title":"Nestru.Encoder.gather_fields_from_struct/2","type":"function"},{"doc":"","ref":"Nestru.Encoder.html#t:t/0","title":"Nestru.Encoder.t/0","type":"type"},{"doc":"","ref":"Nestru.PreDecoder.html","title":"Nestru.PreDecoder","type":"protocol"},{"doc":"Returns a map to be decoded into the struct adopting the Decoder protocol. Nestru calls this function as the first step of the decoding procedure. Useful to adapt the keys of the map to match the field names of the target struct and to specify default values for the missing keys. The first argument is an empty struct value adopting the protocol. The second argument is the context value given to the Nestru.decode/3 function call. The third argument is a map or a binary given to the Nestru.decode/3 function call. If the function returns {:ok, map_or_binary} , then the map_or_binary will be decoded into the struct on the next step using the hint from Nestru.Decoder.decode_fields_hint/3 . If the function returns {:error, message} tuple, then decoding stops, and the error is bypassed to the caller. Any other return value raises an error. The default implementation returns the input value unmodified. To generate the keys translating implementation, set the @derive module attribute to the tuple of Nestru.PreDecoder and the :translate option specifying the map with key names translation. It's useful for copying the value from the input map with the given key to the decodable map with the struct filed key. See the example below. Examples def FruitEnergy do # Keys map can be given with deriving the protocol. # The following will make a function copying the value # of the &quot;energy_factor&quot; key into the :factor key in the map. @derive { Nestru.PreDecoder , translate : %{ &quot;energy_factor&quot; =&gt; :factor } } @derive Nestru.Decoder defstruct [ :value ] end defmodule FruitBox do @derive Nestru.Decoder defstruct [ :items ] defimpl Nestru.PreDecoder do # Put values into the map with the struct&#39;s keys to be decoded later def gather_fields_for_decoding ( _empty_struct , _context , value ) do { :ok , %{ items : Nestru . get ( value , &quot;elements&quot; ) , name : &quot;Default name&quot; } } end end end","ref":"Nestru.PreDecoder.html#gather_fields_for_decoding/3","title":"Nestru.PreDecoder.gather_fields_for_decoding/3","type":"function"},{"doc":"","ref":"Nestru.PreDecoder.html#t:t/0","title":"Nestru.PreDecoder.t/0","type":"type"}]